/*
 * Author: Alexey Mulyukin, 3121
 * Date: 21.03.2014
 */
#define _GNU_SOURCE

#include "main.h"
#include "banking.h"

pid_t parentPid;
FILE* pLogFile;
FILE* pPipeFile;
int childCount;
int processID;

int* pBalance;
int* pLockedBalance;
BalanceHistory* pBalanceHistory;

PipeLineInfo* pPipeLines = NULL;
int pipeLinesWidth = 0;

timestamp_t get_time() {
	return get_physical_time();
}

//------------------------------------------------------------------------------

void initPipeLines(int processCount) {
	pipeLinesWidth = processCount - 1;
	int pipesCount = pipeLinesWidth * processCount;
	pPipeLines = (PipeLineInfo*)malloc(sizeof(PipeLineInfo) * pipesCount);

	for (int id = 0; id < pipesCount; id++) {
		int pipe_ids[2];
		pipe2(pipe_ids, O_NONBLOCK);
		pPipeLines[id].input = pipe_ids[0];
		pPipeLines[id].output = pipe_ids[1];
	}
}

void closePipe(int fd) {
	if (fd != 0) {
		close(fd);
	}
}

int getPipeId(int from, int to) {
	if (to > from) to--;
	return from * pipeLinesWidth + to;
}

int getOpenedPipesFDCount() {
	int pipesCount = pipeLinesWidth * (pipeLinesWidth + 1);
	int result = 0;
	for (int idx = 0; idx < pipesCount; idx++) {
		if (pPipeLines[idx].input)
			result++;
		if (pPipeLines[idx].output)
			result++;
	}

	return result++;
}

void closeUnusedPipes(int selfId) {
	int pipesCount = pipeLinesWidth * (pipeLinesWidth + 1);
	for (int idx = 0; idx < pipesCount; idx++) {
		int fromProcessId = idx / pipeLinesWidth;
		int toProcessId = idx % pipeLinesWidth;
		if (toProcessId >= fromProcessId) toProcessId++;

		if (fromProcessId == selfId) {
			closePipe(pPipeLines[idx].input);
			pPipeLines[idx].input = 0;
		} else if (toProcessId == selfId) {
			closePipe(pPipeLines[idx].output);
			pPipeLines[idx].output = 0;
		} else {
			closePipe(pPipeLines[idx].input);
			pPipeLines[idx].input = 0;
			closePipe(pPipeLines[idx].output);
			pPipeLines[idx].output = 0;
		}
	}

	fprintf(pPipeFile, "Opened for %d PipesFD == %d\n", selfId, getOpenedPipesFDCount());
}

void freePipeLines() {
	if (pPipeLines != NULL) {
		int pipesCount = pipeLinesWidth * (pipeLinesWidth + 1);
		for (int id = 0; id < pipesCount; id++) {
			closePipe(pPipeLines[id].input);
			closePipe(pPipeLines[id].output);
			pPipeLines[id].input = 0;
			pPipeLines[id].output = 0;
		}

		free(pPipeLines);
		pPipeLines = NULL;
	}
}

int writePipe(int fd, const Message* msg) {
	if (fd == 0 || msg == NULL)
		return E_PIPE_INVALID_ARGUMENT;

	write(fd, &msg->s_header, sizeof(MessageHeader));
	if (msg->s_header.s_payload_len > 0) {
		write(fd, msg->s_payload, msg->s_header.s_payload_len);
	}

	return 0;
}

int readPipe(int fd, Message* msg, char isWait) {
	if (fd == 0 || msg == NULL)
		return E_PIPE_INVALID_ARGUMENT;

	while (1) {
		int readed = read(fd, &msg->s_header, sizeof(MessageHeader));

		if (readed < 0) {
			if (isWait) {
				//sleep(1);
			} else {
				return E_PIPE_NO_DATA;
			}
		} else {
			break;
		}
	}

	if (msg->s_header.s_payload_len > 0) {
		int readed = 0;
		do {
			readed = read(fd, msg->s_payload, msg->s_header.s_payload_len);
		} while (readed < 0);
	}

	return 0;
}


//------------------------------------------------------------------------------

/*
 * IPC implementation
 */



/** Send a message to the process specified by id.
 *
 * @param self    Any data structure implemented by students to perform I/O
 * @param dst     ID of recepient
 * @param msg     Message to send
 *
 * @return 0 on success, any non-zero value on error
 */
int send(void * self, local_id dst, const Message * msg) {
	DataInfo* info = (DataInfo*)self;

	int pipeId = pPipeLines[getPipeId(info->senderId, dst)].output;
	return writePipe(pipeId, msg);
}

//------------------------------------------------------------------------------

/** Send multicast message.
 *
 * Send msg to all other processes including parrent.
 * Should stop on the first error.
 *
 * @param self    Any data structure implemented by students to perform I/O
 * @param msg     Message to multicast.
 *
 * @return 0 on success, any non-zero value on error
 */
int send_multicast(void * self, const Message * msg) {
	DataInfo* info = (DataInfo*)self;

	for (int id = 0; id < childCount + 1; id++) {
		if (id != info->senderId) {
			int pipeId = pPipeLines[getPipeId(info->senderId, id)].output;
			if (writePipe(pipeId, msg) < 0)
				return -1;
		}
	}

	return 0;
}

//------------------------------------------------------------------------------


/** Receive a message from the process specified by id.
 *
 * Might block depending on IPC settings.
 *
 * @param self    Any data structure implemented by students to perform I/O
 * @param from    ID of the process to receive message from
 * @param msg     Message structure allocated by the caller
 *
 * @return 0 on success, any non-zero value on error
 */
int receive(void * self, local_id from, Message * msg) {
	DataInfo* info = (DataInfo*)self;

	int pipeId = pPipeLines[getPipeId(from, info->senderId)].input;
	return readPipe(pipeId, msg, 1);
}

//------------------------------------------------------------------------------

/** Receive a message from any process.
 *
 * Receive a message from any process, in case of blocking I/O should be used
 * with extra care to avoid deadlocks.
 *
 * @param self    Any data structure implemented by students to perform I/O
 * @param msg     Message structure allocated by the caller
 *
 * @return 0 on success, any non-zero value on error
 */
int receive_any(void * self, Message * msg) {
	DataInfo* info = (DataInfo*)self;

	for (int id = 0; id <= childCount; id++) {
		if (id != info->senderId) {
			int pipeId = pPipeLines[getPipeId(id, info->senderId)].input;

			if (readPipe(pipeId, msg, 0) == 0) {
				info->receiveId = id;

				return 0;
			}
		}
		//sleep(1);
	}

	return -1;
}

void copy(void* from, void* to, size_t count) {
	for (int i = 0; i < count; i++) {
		((char*)to)[i] = ((char*)from)[i];
	}
}

void zero(void* from, size_t count) {
	for (int i = 0; i < count; i++) {
		((char*)from)[i] = 0;
	}
}

void transfer(void * parent_data, local_id src, local_id dst, balance_t amount)
{
	Message msg;
	DataInfo info;
	info.senderId = PARENT_ID;

	TransferOrder order;
	order.s_src = src;
	order.s_dst = dst;
	order.s_amount = amount;

	msg.s_header.s_local_time = get_time();
	msg.s_header.s_magic = MESSAGE_MAGIC;
	msg.s_header.s_type = TRANSFER;
	msg.s_header.s_payload_len = sizeof(TransferOrder);
	copy(&order, msg.s_payload, msg.s_header.s_payload_len);

	send(&info, src, &msg);

	// wait ACK message
	receive(&info, dst, &msg);
}

/*
void bank_robbery(void * parent_data, local_id max_id)
{
	transfer(parent_data, 1, max_id, 2);
}
*/

//------------------------------------------------------------------------------

void store_history(timestamp_t time, int amount) {
	// find row with current time stamp
	int findIdx = -1;
	for (int idx = 0; idx < pBalanceHistory->s_history_len; idx++) {
		if (pBalanceHistory->s_history[idx].s_time == time) {
			findIdx = idx;
			break;
		}
	}

	BalanceState* pBalanceState;
	if (findIdx < 0 || findIdx >= pBalanceHistory->s_history_len) {
		pBalanceState = &pBalanceHistory->s_history[pBalanceHistory->s_history_len];
		pBalanceHistory->s_history_len++;
	} else {
		pBalanceState = &pBalanceHistory->s_history[findIdx];
	}

	pBalanceState->s_balance = amount;
	pBalanceState->s_balance_pending_in = 0;
	pBalanceState->s_time = time;
}

/*
 * Child workflow
 */

int system_done(pid_t pid, int selfId) {
	// sync
	Message msg;
	sprintf(msg.s_payload, log_done_fmt, get_time(), selfId, pBalance[selfId - 1]);
	msg.s_header.s_magic = MESSAGE_MAGIC;
	msg.s_header.s_payload_len = strlen(msg.s_payload) + 1;
	msg.s_header.s_type = DONE;
	msg.s_header.s_local_time = get_time();

	DataInfo info;
	info.senderId = selfId;

	send_multicast(&info, &msg);

	for (int id = 0; id < childCount + 1; id++) {
		if (id != PARENT_ID && id != selfId) {
			receive(&info, id, &msg);
		}
	}

	store_history(get_time(), pBalance[selfId]);

	// send balance info
	msg.s_header.s_local_time = get_time();
	msg.s_header.s_magic = MESSAGE_MAGIC;
	msg.s_header.s_type = BALANCE_HISTORY;
	msg.s_header.s_payload_len =
			sizeof(pBalanceHistory->s_history_len) +
			sizeof(pBalanceHistory->s_id) +
			sizeof(BalanceState) * (pBalanceHistory->s_history_len);

	copy(pBalanceHistory, msg.s_payload, msg.s_header.s_payload_len);

	send(&info, PARENT_ID, &msg);

	free(pBalanceHistory);

	// sync ended
	fprintf(pLogFile, log_received_all_done_fmt, get_time(), selfId);
	fflush(pLogFile);

	return 0;
}

int system_work(pid_t pid, int selfId) {
	// some work
	char isWorked = 1;

	while (isWorked) {
		Message msg;
		DataInfo info;

		zero(&msg, sizeof(Message));
		zero(&info, sizeof(DataInfo));

		store_history(get_time(), pBalance[selfId]);
		info.senderId = selfId;
		if (receive_any(&info, &msg) == 0) {
			switch (msg.s_header.s_type) {
			case TRANSFER: {
					TransferOrder order;
					zero(&order, sizeof(TransferOrder));

					copy(msg.s_payload, &order, msg.s_header.s_payload_len);
					int amount = order.s_amount;

					if (info.receiveId == PARENT_ID) {
						pBalance[selfId] -= amount;
						store_history(get_time(), pBalance[selfId]);

						send(&info, order.s_dst, &msg);

						fprintf(pLogFile, log_transfer_out_fmt, get_time(), order.s_src, amount, order.s_dst);
						fflush(pLogFile);
					} else {
						fprintf(pLogFile, log_transfer_in_fmt, get_time(), selfId, amount, info.receiveId);
						fflush(pLogFile);

						pBalance[selfId] += amount;
						store_history(get_time(), pBalance[selfId]);

						msg.s_header.s_local_time = get_time();
						msg.s_header.s_magic = MESSAGE_MAGIC;
						msg.s_header.s_type = ACK;
						msg.s_header.s_payload_len = 0;
						send(&info, PARENT_ID, &msg);
					}
				} break;
			case STOP: {
					isWorked = 0;
					store_history(get_time(), pBalance[selfId]);
				} break;
			default:
				break;
			}

			//sleep(1);
		}
	}

	// work is done
	fprintf(pLogFile, log_done_fmt, get_time(), selfId, pBalance[selfId]);
	fflush(pLogFile);

	return system_done(pid, selfId);
}

int system_started(pid_t pid, int selfId) {
	// sync
	fprintf(pLogFile, log_started_fmt, get_time(), selfId, pid, parentPid, pBalance[selfId]);
	fflush(pLogFile);

	Message msg;

	sprintf(msg.s_payload, log_started_fmt, selfId, pid, parentPid);
	msg.s_header.s_magic = MESSAGE_MAGIC;
	msg.s_header.s_payload_len = strlen(msg.s_payload) + 1;
	msg.s_header.s_type = STARTED;
	msg.s_header.s_local_time = get_time();

	DataInfo info;
	info.senderId = selfId;

	send_multicast(&info, &msg);

	for (int id = 0; id < childCount + 1; id++) {
		if (id != PARENT_ID && id != selfId) {
			receive(&info, id, &msg);
		}
	}

	pBalanceHistory = (BalanceHistory*)malloc(sizeof(BalanceHistory));
	pBalanceHistory->s_id = selfId;
	pBalanceHistory->s_history_len = 0;
	store_history(0, pBalance[selfId]);

	// sync complete
	fprintf(pLogFile, log_received_all_started_fmt, get_time(), selfId);
	fflush(pLogFile);

	return system_work(pid, selfId);
}

/*
 * Main process
 */

int main(int argc, char **argv) {
	childCount = 0;

	childCount = atoi(argv[2]);

	if (childCount < 1) {
		return -1;
	}

	pBalance = (int*)malloc(sizeof(int) * childCount + 1);
	pLockedBalance = (int*)malloc(sizeof(int) * childCount + 1);
	pBalance[0] = 0;
	pLockedBalance[0] = 0;
	for (int idx = 1; idx < childCount + 1; idx++) {
		pBalance[idx] = atoi(argv[idx + 2]);
		pLockedBalance[idx] = 0;
	}

	processID = 0;
	parentPid = getpid();

	pLogFile = fopen(evengs_log, "w+");
	pPipeFile = fopen(pipes_log, "w+");

	initPipeLines(childCount + 1);

	pid_t childPid;
	for (int id = 1; id < childCount + 1; id++) {
		childPid = fork();
		if (childPid >= 0) {
			if (childPid == 0) {
				// Child process
				processID = id;
				closeUnusedPipes(id);
				system_started(getpid(), id);
				break;
			} else {
				// Parent process
			}
		} else {
			// Error on fork
		}
	}
	if (childPid != 0) {
		closeUnusedPipes(PARENT_ID);

		DataInfo info;
		Message msg;
		info.senderId = PARENT_ID;

		// Receive all START Messages
		for (int idx = 1; idx <= childCount; idx++) {
			receive(&info, idx, &msg);
		}

		bank_robbery(NULL, childCount);

		// Send STOP messages
		msg.s_header.s_local_time = get_time();
		msg.s_header.s_magic = MESSAGE_MAGIC;
		msg.s_header.s_type = STOP;
		msg.s_header.s_payload_len = 0;

		send_multicast(&info, &msg);

		// Receive all DONE messages;
		for (int idx = 1; idx < childCount + 1; idx++) {
			receive(&info, idx, &msg);
		}

		// Receive Balance History
		AllHistory allHistory;
		allHistory.s_history_len = childCount;
		for (int idx = 1; idx < childCount + 1; idx++) {
			receive(&info, idx, &msg);
			BalanceHistory* pClientHistory = &allHistory.s_history[idx - 1];
			copy(msg.s_payload, pClientHistory, msg.s_header.s_payload_len);
		}

		int status;
		while (wait(&status) > 0);

		print_history(&allHistory);
	}

	freePipeLines();

	free(pBalance);
	free(pLockedBalance);

	fclose(pPipeFile);
	fclose(pLogFile);

	return 0;
}
